package piglets;

import universum.bi.*;
import java.util.TreeSet;
import java.util.concurrent.locks.ReentrantReadWriteLock;

/**
 * Created by IntelliJ IDEA.
 * User: Tanya
 * Date: 02.01.2007
 * Time: 21:31:25
 */
public abstract class SuperHerder {
    public static final float INF   = 1e6f;
    public static final float EPS   = 1e-5f;

    public GladeCell[][] glade;
    public final TreeSet<SuperPiglet> piglets = new TreeSet<SuperPiglet> ();
    public final TreeSet<DrawWell> drawWells = new TreeSet<DrawWell> ();
    private final TreeSet<Integer> knownIds = new TreeSet<Integer> (); 

    private int turnCounter = 0;
    private int turn = 0;

    private final ReentrantReadWriteLock locker = new ReentrantReadWriteLock();
    
    public  final int MAX_TURNS = 200;

    public GameKind gameKind;

    protected void makeTurn(BeingInterface bi) {
        for (SuperPiglet p : piglets) {
            PointInfo curInfo = bi.getPointInfo(p);
            updateWorld(curInfo, p);

            for (PointInfo info : bi.getNeighbourInfo(p)) {
                updateWorld(info, p);
            }
        }
    }

    public void init(UserGameInfo info) {
        this.turn = 0;
        this.turnCounter = 0;
        this.piglets.clear();
        this.knownIds.clear();
        this.drawWells.clear();

        this.gameKind = info.kind;

//        this.MAX_TURNS = 0;

        glade = new GladeCell[Constants.getWidth()][Constants.getHeight()];

        for (int i=0; i< glade.length; ++i) {
            for (int j=0; j< glade[i].length; ++j) {
                glade[i][j] = new GladeCell();
            }
        }
    }

    public final synchronized void updateTurn(BeingInterface bi) {
        if (turnCounter++ == 0) {
            ++turn;
            makeTurn(bi);
        }
        if (turnCounter == piglets.size()) {
            turnCounter = 0;
        }
    }

    public final synchronized int getTurn() {
        return turn;
    }

    public final synchronized void register(SuperPiglet p) {
        piglets.add(p);
        knownIds.add(p.ID);
    }

    public final synchronized void unregister(SuperPiglet p) {
        if (p.ownedWell != null) {
            p.ownedWell.owners.remove(p);
            --turnCounter;
        }
        piglets.remove(p);
    }

    public final void updateWorld(PointInfo info, SuperPiglet p) {
        lockWrite();
        
        Location pos    = info.getLocation();
        GladeCell cell  = glade[pos.getX()][pos.getY()];

        cell.explored   = true;
        cell.growth     = info.getGrowthRate(p);
        cell.count      = info.getCount(p);

        if (cell.growth > 0.f) {
            DrawWell drawWell = new DrawWell(pos, this);
            if (!drawWells.contains(drawWell)) {
                drawWells.add(drawWell);
            }
        }

        // check enemies
        for (Integer id : info.getEntities(p, null)) {
            if (!knownIds.contains(id)) {
                cell.enemy = getTurn();
                break;
            }
        }

        unlockWrite();
    }

    public synchronized GladeCell gladeCellAt(Location pos) {
        return glade[pos.getX()][pos.getY()];
    }

    public void lockRead() {
        locker.readLock().lock();
    }

    public void unlockRead() {
        locker.readLock().unlock();
    }

    public void lockWrite() {
        locker.writeLock().lock();
    }

    public void unlockWrite() {
        locker.writeLock().unlock();
    }
}
